LIBRARY regutil
EXPORTS sysaddrexxmacro
EXPORTS sysclearrexxmacrospace
EXPORTS sysdroprexxmacro
EXPORTS sysloadrexxmacrospace
EXPORTS sysqueryrexxmacro
EXPORTS sysreorderrexxmacro
EXPORTS syssaverexxmacrospace
EXPORTS syscls
EXPORTS syscurpos
EXPORTS syscurstate
EXPORTS sysgetkey
EXPORTS systextscreenread
EXPORTS systextscreensize
EXPORTS syscloseeventsem
EXPORTS sysclosemutexsem
EXPORTS syscreateeventsem
EXPORTS syscreatemutexsem
EXPORTS sysopeneventsem
EXPORTS sysopenmutexsem
EXPORTS sysposteventsem
EXPORTS syspulseeventsem
EXPORTS sysreleasemutexsem
EXPORTS sysrequestmutexsem
EXPORTS sysreseteventsem
EXPORTS syswaiteventsem
EXPORTS sysbootdrive
EXPORTS syswinver
EXPORTS sysos2ver
EXPORTS syslinver
EXPORTS sysversion
EXPORTS sysutilversion
EXPORTS sysdriveinfo
EXPORTS sysdrivemap
EXPORTS syssetpriority
EXPORTS syssleep
EXPORTS sysswitchsession
EXPORTS syssystemdirectory
EXPORTS sysvolumelabel
EXPORTS sysloadfuncs
EXPORTS sysdropfuncs
EXPORTS sysfiledelete
EXPORTS sysfiletree
EXPORTS sysfilesearch
EXPORTS sysmkdir
EXPORTS sysrmdir
EXPORTS syssearchpath
EXPORTS sysfilesystemtype
EXPORTS systempfilename
EXPORTS syscopyobject
EXPORTS sysmoveobject
EXPORTS syscreateshadow
EXPORTS sysstemsort
EXPORTS sysstemcopy
EXPORTS sysstemdelete
EXPORTS syssteminsert
EXPORTS sysdumpvariables
EXPORTS sysqueryprocess
EXPORTS regmultistemsort
EXPORTS regstemread
EXPORTS regstemwrite
EXPORTS regstemdoover
EXPORTS regstemsearch
EXPORTS sysini
EXPORTS syswaitnamedpipe
EXPORTS rxmessagebox
EXPORTS rxwinexec
EXPORTS syssetfiledatetime
EXPORTS sysgetfiledatetime
EXPORTS sysgeterrortext
EXPORTS systounicode
EXPORTS sysfromunicode
EXPORTS syswinencryptfile
EXPORTS syswindecryptfile
EXPORTS syspi
EXPORTS syssqrt
EXPORTS sysexp
EXPORTS syslog
EXPORTS syslog10
EXPORTS syssinh
EXPORTS syscosh
EXPORTS systanh
EXPORTS syspower
EXPORTS syssin
EXPORTS syscos
EXPORTS systan
EXPORTS syscotan
EXPORTS sysarcsin
EXPORTS sysarccos
EXPORTS sysarctan
