;
; regina_os2_dll.def
;
LIBRARY REGINA INITINSTANCE
DESCRIPTION 'Regina Rexx Interpreter'
DATA    MULTIPLE NONSHARED

STACKSIZE   81920

EXPORTS
       RexxAllocateMemory
       RexxDeregisterExit
       RexxDeregisterFunction
       RexxDeregisterSubcom
       RexxFreeMemory
       RexxQueryExit
       RexxQueryFunction
       RexxQuerySubcom
       RexxRegisterExitDll
       RexxRegisterExitExe
       RexxRegisterFunctionDll
       RexxRegisterFunctionExe
       RexxRegisterSubcomDll
       RexxRegisterSubcomExe
       RexxSetHalt
       RexxStart
       RexxVariablePool
       ReginaVersion
       RexxAddMacro
       RexxAddQueue
       RexxClearMacroSpace
       RexxCreateQueue
       RexxDeleteQueue
       RexxDropMacro
       RexxLoadMacroSpace
       RexxPullQueue
       RexxQueryMacro
       RexxQueryQueue
       RexxReorderMacro
       RexxSaveMacroSpace
       RexxCallBack
       ReginaCleanup
       __regina_faked_main
       RexxSetTrace
       RexxResetTrace
