REXXSAA_API {
        global:
                RexxAllocateMemory;
                RexxDeregisterExit;
                RexxDeregisterFunction;
                RexxDeregisterSubcom;
                RexxFreeMemory;
                RexxQueryExit;
                RexxQueryFunction;
                RexxQuerySubcom;
                RexxRegisterExitDll;
                RexxRegisterExitExe;
                RexxRegisterFunctionDll;
                RexxRegisterFunctionExe;
                RexxRegisterSubcomDll;
                RexxRegisterSubcomExe;
                RexxSetHalt;
                RexxStart;
                RexxVariablePool;
        local:
                *;

};
regina_2.0 {
        __regina_faked_main;
        ReginaVersion;
} REXXSAA_API;
regina_2.2 {
        RexxAddMacro;
        RexxAddQueue;
        RexxClearMacroSpace;
        RexxCreateQueue;
        RexxDeleteQueue;
        RexxDropMacro;
        RexxLoadMacroSpace;
        RexxPullQueue;
        RexxQueryMacro;
        RexxQueryQueue;
        RexxReorderMacro;
        RexxSaveMacroSpace;
} REXXSAA_API;
regina_3.1 {
        RexxCallBack;
} REXXSAA_API;
regina_3.3 {
        ReginaCleanup;
} REXXSAA_API;
regina_3.7 {
        RexxSetTrace;
        RexxResetTrace;
} REXXSAA_API;
