#include "winver.h"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 3,0,0,0
 PRODUCTVERSION 3,0,0,0
 FILEFLAGSMASK 0x0L
 FILEFLAGS 0x0L
 FILEOS 0x40004L
 FILETYPE VFT_APP
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Author", "Mark Hessling\0"
            VALUE "FileDescription", "Regina Rexx Tokeniser\0"
            VALUE "FileVersion", "3.0\0"
            VALUE "InternalName", "tRegina\0"
            VALUE "LegalCopyright", "Copyright © 2002\0"
            VALUE "License", "GNU Library General Public License. See file COPYING-LIB for details.\0"
            VALUE "OriginalFilename", "TREGINA.EXE\0"
            VALUE "ProductName", "Regina Rexx Interpreter\0"
            VALUE "ProductVersion", "3.0\0"
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

2 ICON    DISCARDABLE     "reginaw32.ico"
