LIBRARY DLL
DESCRIPTION 'regina'
EXPORTS RexxDeregisterExit
EXPORTS RexxDeregisterFunction
EXPORTS RexxDeregisterSubcom
EXPORTS RexxQueryFunction
EXPORTS RexxQuerySubcom
EXPORTS RexxRegisterExitExe
EXPORTS RexxRegisterExitDll
EXPORTS RexxQueryExit
EXPORTS RexxRegisterFunctionDll
EXPORTS RexxRegisterFunctionExe
EXPORTS RexxRegisterSubcomDll
EXPORTS RexxRegisterSubcomExe
EXPORTS RexxStart
EXPORTS RexxSetHalt
EXPORTS RexxVariablePool
EXPORTS RexxAllocateMemory
EXPORTS RexxFreeMemory
EXPORTS ReginaVersion
EXPORTS __regina_faked_main
EXPORTS RexxAddMacro
EXPORTS RexxAddQueue
EXPORTS RexxClearMacroSpace
EXPORTS RexxCreateQueue
EXPORTS RexxDeleteQueue
EXPORTS RexxDropMacro
EXPORTS RexxLoadMacroSpace
EXPORTS RexxPullQueue
EXPORTS RexxQueryMacro
EXPORTS RexxQueryQueue
EXPORTS RexxReorderMacro
EXPORTS RexxSaveMacroSpace
EXPORTS RexxCallBack
EXPORTS ReginaCleanup
EXPORTS RexxSetTrace
EXPORTS RexxResetTrace
